# -*- encoding: utf-8 -*-
"""
@文件        :GmAppCommand.py
@说明        :
@时间        :2024/11/27 18:37:49
@作者        :Zack
@版本        :1.0
"""

from cleo.helpers import argument, option

from gmscaffold.base.command import BaseCommand, Command
from gmscaffold.entities.GmApp import GmAppData
from gmscaffold.utils import signal as _signal


class RemoveAppCommand(BaseCommand):
    """ """

    name = "gm_app remove_app"
    description = "Remove applications"
    arguments = [
        argument(
            "name",
            description="Application name?",
            optional=True,
        )
    ]
    options = [
        option(
            "remove_app",
            description="Remove gm app",
            flag=True,
        )
    ]

    def remove_app(self, gmapp_name: str = None):
        """Remove applications"""
        print("remove_app", gmapp_name)

    def handle(self):
        self.remove_app()


class CreateAppCommand(BaseCommand):
    """ """

    name = "gm_app create_app"
    description = "Create Applications"

    def create_app(self, *args, **kwargs):
        """Create gm applications"""
        gmapp_data = GmAppData()
        self._input_project_name(gmapp_data)._input_project_author(
            gmapp_data
        )._input_project_email(gmapp_data)._input_project_version(
            gmapp_data
        )._input_project_desc(
            gmapp_data
        )._input_project_date(
            gmapp_data
        )
        self.Gm._signals.send(signal=_signal.create_gmapp_project, data=gmapp_data)

    def handle(self):
        self.create_app()


class GmAppCommand(BaseCommand):
    """ """

    name = "gm_app"
    description = "GM applications operator"

    commands = [RemoveAppCommand, CreateAppCommand]



# -*- encoding: utf-8 -*-
"""
@文件        :exceptions.py
@说明        :
@时间        :2024/11/25 11:55:00
@作者        :Zack
@版本        :1.0
"""


class BaseGmException(Exception):
    """ """


class NoDataError(BaseGmException):
    """ """


class BaseCommandException(BaseGmException):
    """ """


class BuildProjectCommandError(BaseCommandException):
    """ """


class BuildValueError(BaseGmException):
    """ """



# -*- encoding: utf-8 -*-
"""
@文件        :web.py
@说明        :
@时间        :2024/11/27 18:45:12
@作者        :Zack
@版本        :1.0
"""

from gmscaffold.base.command import BaseBuilder
from gmscaffold.entities.web import WebData
from gmscaffold.utils import builder
from gmscaffold.utils import signal as _signal
from gmscaffold.utils import template


class Web(BaseBuilder):
    """ """

    component_name = "sanic"

    def __init__(self, Gm):
        """ """
        self.Gm = Gm
        signals_callbacks = {
            _signal.create_web_project: self.start_project,
            _signal.remove_web_project: self.remove_project,
            _signal.create_web_app: self.start_app,
            _signal.remove_web_app: self.remove_app,
        }

        super().__init__(signals_callbacks)

    def start_project(self, sender, **kwargs):
        """ """
        self.check_data(kwargs)
        project_item: WebData = kwargs["data"]
        tpl_data = builder.make_template_data(self, project_item)

        template.build_project(tpl_data)

    def remove_project(self):
        """ """

    def start_app(self, sender, **kwargs):
        """ """
        self.check_data(kwargs)
        project_item: WebData = kwargs["data"]
        tpl_data = builder.make_template_app(self, project_item)

        template.build_app(tpl_data)

    def remove_app(self):
        """ """



# -*- encoding: utf-8 -*-
"""
@文件        :command.py
@说明        :
@时间        :2024/11/25 13:57:46
@作者        :Zack
@版本        :1.0
"""
import datetime
import subprocess
from abc import ABCMeta
from dataclasses import asdict
from typing import Any, Dict, Optional, TypeVar

from cleo.commands.command import Command

from gmscaffold.base.dataclass import BaseData, TemplateData
from gmscaffold.exceptions import NoDataError
from gmscaffold.utils import path as _path
from gmscaffold.utils import signal as _signal
from gmscaffold.utils.conf import get_module_name


class _ProjectObject:
    """ """

    name: Optional[str]
    author: Optional[str]
    email: Optional[str]
    description: Optional[str]
    version: Optional[str]
    date: Optional[str]


ProjectObject = TypeVar("ProjectObject", bound=_ProjectObject)


class BaseCommand(Command, metaclass=ABCMeta):
    """ """

    def __init__(self, Gm):
        """ """
        self.Gm = Gm
        super().__init__()

    def _input_project_name(self, project_object: ProjectObject):
        """ """
        project_name = input("请输入项目名称: ")
        project_object.name = project_name
        return self

    def _input_project_author(self, project_object: ProjectObject):
        """ """
        author_name = input("请输入作者名称: ")
        project_object.author = author_name
        return self

    def _input_project_email(self, project_object: ProjectObject):
        """ """
        email = input("请输入联系邮箱：")
        project_object.email = email
        return self

    def _input_project_desc(self, project_object: ProjectObject):
        """ """
        description = input("请输入项目描述信息：")
        project_object.description = description
        return self

    def _input_project_version(self, project_object: ProjectObject):
        """ """
        version = input("请输入项目版本号：")
        project_object.version = version
        return self

    def _input_project_date(self, project_object: ProjectObject):
        """ """
        project_object.date = datetime.datetime.now().strftime("%Y/%m/%d %H:%M:%S")
        return self

    def check_call(self, *args, **kwargs):
        """Execute the command"""
        return subprocess.check_call(*args, **kwargs)


class BaseBuilder(object):
    """ """

    def __init__(self, signals_callbacks):
        """ """
        self.register_signals(signals_callbacks)

    def check_data(self, kwargs: Dict[str, Any]):
        """ """
        if not kwargs.get("data"):
            raise NoDataError("No data to create spider project")

    def register_signals(self, signals_callbacks):
        """ """
        for signal, callback in signals_callbacks.items():
            self.Gm._signals.connect(callback, signal)

    def open_signals(self, signals_callbacks):
        """ """
        for signal in signals_callbacks:
            """ """
            self.Gm._signals.send(signal, data=_signal.OPEN_CONNECT)

    @classmethod
    def from_settings(cls, settings):
        """ """
        return cls(settings)



# -*- encoding: utf-8 -*-
"""
@文件        :command.py
@说明        :
@时间        :2024/11/26 14:57:46
@作者        :Zack
@版本        :1.0
"""

import logging
from functools import partial

_LOG_FORMAT = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"


class LoggerWriter:
    """ """

    def __init__(self, logger, level):
        """ """
        self.logger = logger
        self.level = level

    def write(self, message):
        if message != "\n":  # avoid printing bare newlines, if you like
            self.logger.log(self.level, message)

    def flush(self):
        # doesn't actually do anything, but might be expected of a file-like
        # object - so optional depending on your situation
        pass

    def close(self):
        # doesn't actually do anything, but might be expected of a file-like
        # object - so optional depending on your situation. You might want
        # to set a flag so that later calls to write raise an exception
        pass


class Logger:
    """ """

    def __init__(self, level=logging.DEBUG, **kwargs):
        """
        Do basic configuration for the logging system.

        This function does nothing if the root logger already has handlers
        configured, unless the keyword argument *force* is set to ``True``.
        It is a convenience method intended for use by simple scripts
        to do one-shot configuration of the logging package.

        The default behaviour is to create a StreamHandler which writes to
        sys.stderr, set a formatter using the BASIC_FORMAT format string, and
        add the handler to the root logger.

        A number of optional keyword arguments may be specified, which can alter
        the default behaviour.

        filename  Specifies that a FileHandler be created, using the specified
                filename, rather than a StreamHandler.
        filemode  Specifies the mode to open the file, if filename is specified
                (if filemode is unspecified, it defaults to 'a').
        format    Use the specified format string for the handler.
        datefmt   Use the specified date/time format.
        style     If a format string is specified, use this to specify the
                type of format string (possible values '%', '{', '$', for
                %-formatting, :meth:`str.format` and :class:`string.Template`
                - defaults to '%').
        level     Set the root logger level to the specified level.
        stream    Use the specified stream to initialize the StreamHandler. Note
                that this argument is incompatible with 'filename' - if both
                are present, 'stream' is ignored.
        handlers  If specified, this should be an iterable of already created
                handlers, which will be added to the root handler. Any handler
                in the list which does not have a formatter assigned will be
                assigned the formatter created in this function.
        force     If this keyword  is specified as true, any existing handlers
                attached to the root logger are removed and closed, before
                carrying out the configuration as specified by the other
                arguments.
        encoding  If specified together with a filename, this encoding is passed to
                the created FileHandler, causing it to be used when the file is
                opened.
        errors    If specified together with a filename, this value is passed to the
                created FileHandler, causing it to be used when the file is
                opened in text mode. If not specified, the default value is
                `backslashreplace`.

        Note that you could specify a stream created using open(filename, mode)
        rather than passing the filename and mode in. However, it should be
        remembered that StreamHandler does not close its stream (since it may be
        using sys.stdout or sys.stderr), whereas FileHandler closes its stream
        when the handler is closed.

        .. versionchanged:: 3.2
        Added the ``style`` parameter.

        .. versionchanged:: 3.3
        Added the ``handlers`` parameter. A ``ValueError`` is now thrown for
        incompatible arguments (e.g. ``handlers`` specified together with
        ``filename``/``filemode``, or ``filename``/``filemode`` specified
        together with ``stream``, or ``handlers`` specified together with
        ``stream``.

        .. versionchanged:: 3.8
        Added the ``force`` parameter.

        .. versionchanged:: 3.9
        Added the ``encoding`` and ``errors`` parameters.
        """
        logging.basicConfig(level=level)
        self.logger = logging.getLogger("GmScaffold")

    @classmethod
    def from_log_file(cls, log_file: str, level=logging.DEBUG, formatter=None):
        """ """
        fh = logging.FileHandler(log_file)
        formatter = logging.Formatter(formatter or _LOG_FORMAT)
        fh.setFormatter(formatter)
        fh.setLevel(level)
        logger = cls(level)
        logger.add_handle(fh)
        return logger

    def add_handle(self, fh):
        """ """
        self.logger.addHandler(fh)

    def io_logger(self, logger, mode) -> LoggerWriter:
        """ """
        return LoggerWriter(logger, mode)

    def io_partial(self, mode):
        """ """
        return partial(self.io_logger, logger=self.logger, mode=mode)

    def debug(self, message: str):
        """ """
        print(message, file=self.io_logger(logger=self.logger, mode=logging.DEBUG))

    def info(self, message: str):
        """ """
        print(message, file=self.io_logger(logger=self.logger, mode=logging.INFO))

    def warn(self, message: str):
        """ """
        print(message, file=self.io_logger(logger=self.logger, mode=logging.WARN))

    def error(self, message: str):
        """ """
        print(message, file=self.io_logger(logger=self.logger, mode=logging.ERROR))

    def critical(self, message: str):
        """ """
        print(message, file=self.io_logger(logger=self.logger, mode=logging.CRITICAL))


logger = Logger()



# -*- encoding: utf-8 -*-
"""
@文件        :GmApp.py
@说明        :
@时间        :2024/11/27 18:40:17
@作者        :Zack
@版本        :1.0
"""
from dataclasses import dataclass

from gmscaffold.base.dataclass import BaseData


@dataclass
class GmAppData(BaseData):
    """ """



from typing import Any

from gmscaffold.dispatcher import dispatcher


class SignalManager:
    """ """

    _instance = None

    def __new__(cls, *args, **kwargs):
        """ """
        if cls._instance is None:
            cls._instance = super().__new__(cls, *args, **kwargs)
        return cls._instance

    def __init__(self, sender: Any = dispatcher.Anonymous):
        """ """
        self.sender: Any = sender

    def connect(self, receiver: Any, signal: Any, **kwargs: Any) -> None:
        """
        Connect a receiver function to a signal.

        The signal can be any object, although Scrapy comes with some
        predefined signals that are documented in the :ref:`topics-signals`
        section.

        :param receiver: the function to be connected
        :type receiver: collections.abc.Callable

        :param signal: the signal to connect to
        :type signal: object
        """
        kwargs.setdefault("sender", self.sender)
        dispatcher.connect(receiver, signal, **kwargs)

    def disconnect(self, receiver: Any, signal: Any, **kwargs: Any) -> None:
        """
        Disconnect a receiver function from a signal. This has the
        opposite effect of the :meth:`connect` method, and the arguments
        are the same.
        """
        kwargs.setdefault("sender", self.sender)
        dispatcher.disconnect(receiver, signal, **kwargs)

    def disconnect_all(self, signal: Any, **kwargs: Any) -> None:
        """
        Disconnect all receivers from the given signal.

        :param signal: the signal to disconnect from
        :type signal: object
        """
        kwargs.setdefault("sender", self.sender)
        """Disconnect all signal handlers. Useful for cleaning up after running
        tests
        """
        for receiver in dispatcher.liveReceivers(dispatcher.getAllReceivers(self.sender, signal)):
            dispatcher.disconnect(receiver, signal=signal, sender=self.sender)

    def send(self, signal=Any, *arguments, **kwargs):
        """Send signal from sender to all connected receivers.

        signal -- (hashable) signal value, see connect for details

        sender -- the sender of the signal

            if Any, only receivers registered for Any will receive
            the message.

            if Anonymous, only receivers registered to receive
            messages from Anonymous or Any will receive the message

            Otherwise can be any python object (normally one
            registered with a connect if you actually want
            something to occur).

        arguments -- positional arguments which will be passed to
            *all* receivers. Note that this may raise TypeErrors
            if the receivers do not allow the particular arguments.
            Note also that arguments are applied before named
            arguments, so they should be used with care.

        named -- named arguments which will be filtered according
            to the parameters of the receivers to only provide those
            acceptable to the receiver.

        Return a list of tuple pairs [(receiver, response), ... ]

        if any receiver raises an error, the error propagates back
        through send, terminating the dispatch loop, so it is quite
        possible to not have all receivers called if a raises an
        error.
        """
        kwargs.setdefault("sender", self.sender)
        dispatcher.send(signal=signal, *arguments, **kwargs)



# -*- encoding: utf-8 -*-
"""
@文件        :SpiderCommands.py
@说明        :
@时间        :2024/11/26 18:51:23
@作者        :Zack
@版本        :1.0
"""

from collections import defaultdict
from typing import Any


class _Parameter:
    """Used to represent default parameter values."""

    def __repr__(self):
        return self.__class__.__name__


class _Anonymous(_Parameter):
    """Singleton used to signal "Anonymous Sender"

    The Anonymous object is used to signal that the sender
    of a message is not specified (as distinct from being
    "any sender").  Registering callbacks for Anonymous
    will only receive messages sent without senders.  Sending
    with anonymous will only send messages to those receivers
    registered for Any or Anonymous.

    Note:
        The default sender for connect is Any, while the
        default sender for send is Anonymous.  This has
        the effect that if you do not specify any senders
        in either function then all messages are routed
        as though there was a single sender (Anonymous)
        being used everywhere.
    """


Anonymous = _Anonymous()


class dispatcher(object):
    """ """

    Anonymous = _Anonymous()
    handlers = defaultdict(list)

    @classmethod
    def connect(cls, receiver: Any, signal: Any, **kwargs: Any):
        """ """
        cls.handlers[signal].append(receiver)

    @classmethod
    def send(cls, signal=Any, *arguments, **kwargs):
        """ """
        for handler in cls.handlers[signal]:
            handler(*arguments, **kwargs)

    @classmethod
    def disconnect(cls, receiver, signal: Any):
        """ """
        raise NotImplementedError

    @classmethod
    def disconnect_all(cls, signal: Any, **kwargs: Any):
        """ """
        raise NotImplementedError

    @classmethod
    def liveReceivers(cls, signal: Any):
        """ """
        raise NotImplementedError

    @classmethod
    def getAllReceivers(cls, sender=Any, signal=Any):
        """ """
        raise NotImplementedError



# -*- encoding: utf-8 -*-
"""
@文件        :GmApp.py
@说明        :
@时间        :2024/11/27 18:40:46
@作者        :Zack
@版本        :1.0
"""
from dataclasses import asdict

from gmscaffold.base.command import BaseBuilder
from gmscaffold.base.dataclass import TemplateData
from gmscaffold.entities.GmApp import GmAppData
from gmscaffold.log import logger
from gmscaffold.utils import path as _path
from gmscaffold.utils import signal as _signal
from gmscaffold.utils import template
from gmscaffold.utils.conf import get_module_name


class GmApp(BaseBuilder):
    """ """

    component_name = "gmapp"

    def __init__(self, Gm):
        """ """
        self.Gm = Gm

        signals_callbacks = {
            _signal.create_gmapp_project: self.create_gm_app,
            _signal.remove_gmapp_project: self.remove_gm_app,
        }

        super().__init__(signals_callbacks)

    def create_gm_app(self, sender, **kwargs):
        """ """
        self.check_data(kwargs)
        project_item: GmAppData = kwargs["data"]
        module_name = get_module_name(self)
        module_tpl = self.Gm._settings.extension_template(module_name)
        files_list = _path.get_files_in_directory(module_tpl)
        tpl_root = project_item.name
        project_data = asdict(project_item)
        tpl_data = TemplateData(files_list, project_data, module_tpl, tpl_root)

        # TODO: 测试
        # test_folder = os.path.dirname(os.path.dirname(__file__))
        # tpl_data.target_dir = os.path.join(test_folder, "test")

        template.build_project(tpl_data)
        logger.debug(f"[{project_item.name}] create_gm_app successfully")

    def remove_gm_app(self, sender, **kwargs):
        """ """
        # TODO: NotImplementedError
        print("[GmApp] remove_gm_app...", sender, kwargs)

    @classmethod
    def from_settings(cls, settings):
        """ """
        return cls(settings)



# -*- encoding: utf-8 -*-
"""
@文件        :test_cmdline.py
@说明        :
@时间        :2024/11/25 13:40:03
@作者        :Zack
@版本        :1.0
"""
import os
import sys
import unittest

sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.dirname(__file__))))

from gmscaffold.cmdline import GmScaffoldMain
from gmscaffold.commands.SpiderCommand import SpiderCommand
from gmscaffold.log import Logger
from gmscaffold.settings.default_settings import STATIC_DIR
from gmscaffold.utils.path import *
from gmscaffold.utils.template import *


class TestSpider(unittest.TestCase):
    """ """

    @unittest.skip
    def test_get_dirs(self):
        """ """
        directory = STATIC_DIR
        dirs = get_files_in_directory(directory)
        print("[*] dirs >>> ", dirs)

    @unittest.skip
    def test_spider_template(self):
        """ """
        test_dir = os.getcwd()
        tmp = Template(test_dir)
        print("[*] test_spider_template >> ", tmp.to_cwd("logs"))

    @unittest.skip
    def test_spider(self):
        """ """
        sc = SpiderCommand()
        sc.create_spider()

    @unittest.skip
    def test_loader_web(self):
        """ """
        gm = GmScaffoldMain()
        gm.web("start_project")

    # @unittest.skip
    def test_loader(self):
        """ """
        gm = GmScaffoldMain()
        gm.spider("create_spider")
        # build_component_list()
        # pl = ProcessLoader()
        # pl.load_extensions()

    @unittest.skip
    def test_logger(self):
        """ """
        file = os.path.join(os.path.dirname(os.path.dirname(__file__)), "logs", "test.log")
        logger = Logger.from_log_file(file)
        logger.info("ddddddddddddd")


if __name__ == "__main__":
    """ """
    unittest.main()



# -*- encoding: utf-8 -*-
"""
@文件        :conf.py
@说明        :
@时间        :2024/11/25 13:33:13
@作者        :Zack
@版本        :1.0
"""
import os
from configparser import ConfigParser
from pathlib import Path
from typing import Iterator, List

from gmscaffold.settings import Settings


def build_component_list(Gm):
    """ """
    from gmscaffold.loader.process import ProcessLoader

    Gm._from_settings(Settings())
    return ProcessLoader(Gm)


def get_config(config_path: str = None) -> ConfigParser:
    """Get Scrapy config file as a ConfigParser"""
    sources = get_sources(config_path)
    cfg = ConfigParser()
    cfg.read(sources)
    return cfg


def get_sources(config_path: str = None) -> List[str]:
    xdg_config_home = os.environ.get("GM_CONFIG_HOME") or Path("~/.config").expanduser()
    sources = [
        str(Path(xdg_config_home) / "gm.cfg"),
    ]
    if config_path:
        sources.append(config_path)
    return sources


def get_extensions_path():
    """ """
    extensions_path = os.path.join(os.path.dirname(os.path.dirname(__file__)), "extensions")
    return extensions_path


def gen_list_dir(base_dir) -> Iterator:
    """ """
    yield from os.listdir(base_dir)


def get_module_name(module: object):
    """ """
    ext_name = "{0.__class__.__module__}.{0.__class__.__name__}".format(module)
    return ext_name


def get_module_basename_with_path(path: str):
    """ """
    return os.path.basename(path)



# -*- encoding: utf-8 -*-
"""
@文件        :cmdline.py
@说明        :
@时间        :2024/11/25 10:54:59
@作者        :Zack
@版本        :1.0
"""
from __future__ import annotations

import os
import sys
from importlib import import_module
from typing import TYPE_CHECKING

import pyfiglet
from cleo.application import Application
from cleo.exceptions import CleoLogicError
from cleo.loaders.factory_command_loader import FactoryCommandLoader

sys.path.insert(0, os.path.dirname(os.path.dirname(__file__)))

from gmscaffold.signalmanager import SignalManager
from gmscaffold.utils.conf import build_component_list

application = Application()


if TYPE_CHECKING:
    from collections.abc import Callable

    from cleo.commands.command import Command


class CommandLoader(FactoryCommandLoader):
    def register_factory(
        self, command_name: str, factory: Callable[[], Command]
    ) -> None:
        if command_name in self._factories:
            raise CleoLogicError(f'The command "{command_name}" already exists.')

        self._factories[command_name] = factory


def load_command(name: str, self) -> Callable[[], Command]:
    def _load() -> Command:
        package, cls = COMMANDS[name].rsplit(".", maxsplit=1)
        module = import_module(package)
        command_class = getattr(module, cls)
        command: Command = command_class(self)
        return command

    return _load


COMMANDS = {
    # gm_app commands
    "gm_app create_app": "gmscaffold.commands.GmAppCommand.CreateAppCommand",
    # "gm_app remove_app": "gmscaffold.commands.GmAppCommand.RemoveAppCommand",
    # spider commands
    "spider create_spider": "gmscaffold.commands.SpiderCommand.CreateSpiderCommand",
    # "spider remove_spider": "gmscaffold.commands.SpiderCommand.RemoveSpiderCommand",
    # web commands
    "web start_project": "gmscaffold.commands.WebCommand.CreateWebCommand",
    "web remove_project": "gmscaffold.commands.WebCommand.RemoveWebCommand",
    "web start_app": "gmscaffold.commands.WebCommand.StartAppWebCommand",
    "web remove_app": "gmscaffold.commands.WebCommand.RemoveAppWebCommand",
}


class GmScaffoldMain(Application):
    """ """

    def __init__(self):
        """ """
        self._signals = SignalManager()
        build_component_list(self)
        super().__init__()
        command_loader = CommandLoader(
            {name: load_command(name, self) for name in COMMANDS}
        )
        self.set_command_loader(command_loader)

    def _from_settings(self, settings):
        """ """
        self._settings = settings

    @property
    def command_loader(self) -> CommandLoader:
        command_loader = self._command_loader
        assert isinstance(command_loader, CommandLoader)
        return command_loader


def execute(argv=None, settings=None):
    """命令行交互处理"""
    welcome_txt = pyfiglet.figlet_format("GMSSH", font="lean")
    print(welcome_txt)
    GmScaffoldMain().run()


if __name__ == "__main__":

    execute()



# -*- encoding: utf-8 -*-
"""
@文件        :defer.py
@说明        :
@时间        :2024/11/25 11:08:37
@作者        :Zack
@版本        :1.0
"""
from __future__ import annotations

from typing import Awaitable, TypeVar

_SelfResultT = TypeVar("_SelfResultT")


class Deferred(Awaitable[_SelfResultT]):
    """ """
